# This file is part of CovidPy v0.0.2.
#
# The project has been distributed in the hope it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You can use it and/or modify it under the terms of the GNU General Public License v3.0 or later.
# You should have received a copy of the GNU General Public License along with the project.

import requests
import os
import qrcode
import zlib
import cbor2
import pyzbar.pyzbar
from base45 import b45encode, b45decode
from cose.algorithms import Es256
from cose.keys.curves import P256
from cose.algorithms import Es256
from cose.keys.keyparam import KpKty, KpAlg, EC2KpD, EC2KpCurve
from cose.headers import Algorithm, KID
from cose.keys import CoseKey
from cose.keys.keyparam import KpAlg, EC2KpD, EC2KpCurve
from cose.keys.keyparam import KpKty
from cose.keys.keytype import KtyEC2
from cose.messages import Sign1Message
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import load_pem_private_key
import PIL

from .verifier import DCCVerifier
from .types import QRCode
from .errors import InvalidDCC

class CovidPy:
    def __init__(self, disable_keys_update:bool=False, disable_blalcklist_update:bool=False, disable_blacklist:bool=False) -> None:
        self.__autoblacklist = disable_blalcklist_update
        self.__autokids = disable_keys_update
        self.__disableblacklist = disable_blacklist
        self.__verifier = DCCVerifier(self.__autoblacklist, self.__autokids)
        self.__verifier.load_eu_keys()
        if not self.__disableblacklist:
            self.__verifier.load_blacklist()
        self.certspath = 'certs'

    def __decodecertificate(self, cert):
        img = PIL.Image.open(cert) if isinstance(cert, str) else cert.to_bytesio
        data = pyzbar.pyzbar.decode(img)
        cert = data[0].data.decode()
        if cert.startswith('HC1:'):
            b45data = cert.replace("HC1:", "")
            compresseddata = b45decode(b45data)
            decompressed = zlib.decompress(compresseddata)
            return decompressed
        else:
            raise InvalidDCC('The given code is not a DCC, check the \'details\' attribute for more details', 'HC1_MISSING')
    
    def __getUVCI(self,dictionary):
        for key, value in dictionary.items():
            if isinstance(value, dict):
                if (val := self.__getUVCI(value)) is not None:
                    return val
            elif isinstance(value,list):
                for v in value:
                    if isinstance(v, dict):
                        ci = v.get('ci', None)
                        return ci
            else:
                if key == "ci":
                    return value
        else:
            return None
    
    def __is_blacklisted(self, raw:dict):
        ci = self.__getUVCI(raw)
        return ci in self.verifier.blacklist

            
    def decode(self, cert): 
        cbordata = self.__decodecertificate(cert)
        decoded = cbor2.loads(cbordata)
        return cbor2.loads(decoded.value[2]) 

    def __finddatapaths(self): #TODO upload to pypi and remove this function
        rootdir = os.getcwd()

        for subdir, _, files in os.walk(rootdir):
            for file in files:
                filepath = subdir + os.sep + file

                if filepath.endswith("lib.py"):
                    return subdir + os.sep + 'certs'

    def __genqr(self, payload:dict):
        cbordata = cbor2.dumps(payload)
        with open(f'{self.certspath}\\dsc-worker.pem', "rb") as file:
            pem = file.read()
            cert = x509.load_pem_x509_certificate(pem)
            fingerprint = cert.fingerprint(hashes.SHA256())
            keyid = fingerprint[0:8]

        with open(f'{self.certspath}\\dsc-worker.key', "rb") as file:
            pem = file.read()
            keyfile = load_pem_private_key(pem, password=None)
            priv = keyfile.private_numbers().private_value.to_bytes(32, byteorder="big")

        msg = Sign1Message(phdr={Algorithm: Es256, KID: keyid}, payload=cbordata)

        cose_key = {
        KpKty: KtyEC2,
        KpAlg: Es256, 
        EC2KpCurve: P256, 
        EC2KpD: priv,
        }

        msg.key = CoseKey.from_dict(cose_key)
        out = msg.encode()

        out = zlib.compress(out, 9)

        out = b'HC1:' + b45encode(out).encode('ascii')

        return qrcode.make(out), out

    def encode(self, data:dict)->QRCode:
        gqr = self.__genqr(data)
        qr = QRCode(gqr[0], gqr[1],self.__is_blacklisted(data), self)
        return qr
    def verify(self, cert):
        if not self.disableblacklist and not self.__is_blacklisted(self.decode(cert)):
            return self.__verifier.is_valid(self.__decodecertificate(cert))
        else:
            return False


# This file is part of CovidPy v0.0.2.
#
# The project has been distributed in the hope it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You can use it and/or modify it under the terms of the GNU General Public License v3.0 or later.
# You should have received a copy of the GNU General Public License along with the project.

from .lib import CovidPy
from .types import QRCode
from .verifier import DCCVerifier
from .errors import InvalidDCC


# This file is part of CovidPy v0.0.2.
#
# The project has been distributed in the hope it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You can use it and/or modify it under the terms of the GNU General Public License v3.0 or later.
# You should have received a copy of the GNU General Public License along with the project.

from base64 import b64encode, b64decode
from cose.headers import KID
import requests, json
from cose.algorithms import Es256
from cose.keys.curves import P256
from cose.algorithms import Es256, Ps256
from cose.keys import CoseKey
from cose.keys.keyparam import KpAlg, EC2KpX, EC2KpY, EC2KpCurve, RSAKpE, RSAKpN
from cose.keys.keyparam import KpKty
from cose.keys.keytype import KtyEC2, KtyRSA
from cose.messages import CoseMessage
from cryptography.utils import int_to_bytes
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives import serialization
import schedule

class DCCVerifier:
    def __init__(self, disblrefresh, diskidsrefresh):
        self.diskidsrefresh = diskidsrefresh
        self.disblrefresh = disblrefresh
        self.kids:dict = {}
        self.sched = None
        self.blacklist:list = []
        if not self.disblrefresh and not self.diskidsrefresh:
            self.sched = schedule.every().day.at("08:00").do(self.reload_all)
        elif self.disblrefresh and not self.diskidsrefresh:
            self.sched = schedule.every().day.at("08:00").do(self.reload_kids)
        elif self.diskidsrefresh and not self.disblrefresh:
            self.sched = schedule.every().day.at("08:00").do(self.reload_bl)

    def reload_all(self):
        self.reload_kids()
        self.reload_bl()

    def reload_bl(self):
        self.blacklist = []
        self.load_blacklist()
    
    def reload_kids(self):
        self.kids = {}
        self.load_eu_keys()

    def append_kid(self, kid_b64, key_b64):
        asn1data = b64decode(key_b64)

        pub = serialization.load_der_public_key(asn1data)
        if (isinstance(pub, RSAPublicKey)):
              self.kids[kid_b64] = CoseKey.from_dict(
               {
                    KpAlg: Ps256, 
                    KpKty: KtyRSA,
                    RSAKpE: int_to_bytes(pub.public_numbers().e),
                    RSAKpN: int_to_bytes(pub.public_numbers().n)
               })
        elif (isinstance(pub, EllipticCurvePublicKey)):
              self.kids[kid_b64] = CoseKey.from_dict(
               {
                    EC2KpCurve: P256, 
                    KpKty: KtyEC2, 
                    KpAlg: Es256,  
                    EC2KpX: pub.public_numbers().x.to_bytes(32, byteorder="big"),
                    EC2KpY: pub.public_numbers().y.to_bytes(32, byteorder="big")
               })
        else:
            pass #INVALID TYPE 

    def load_blacklist(self):
        settings = requests.get('https://get.dgc.gov.it/v1/dgc/settings').json()
        for sett in settings:
            if sett.get('name',None) == 'black_list_uvci':
                self.blacklist = sett.get('value').split(';')
                break

    def load_eu_keys(self):
        keys = requests.get('https://verifier-api.coronacheck.nl/v4/verifier/public_keys')
        kjson = keys.json()
        payload = b64decode(kjson['payload'])
        trustlist = json.loads(payload)
        eutrusts = trustlist['eu_keys']
        for b64kid in eutrusts:
            self.append_kid(b64kid,eutrusts[b64kid][0]['subjectPk'])
    def is_valid(self, key):
        cose = CoseMessage.decode(key)
        a_kid = cose.phdr[KID] if KID in cose.phdr.keys() else cose.uhdr[KID]
        b64_a_kid = b64encode(a_kid).decode('ASCII')
        if not b64_a_kid in self.kids:
            return False
        else:
            key  = self.kids[b64_a_kid]

            cose.key = key
            if not cose.verify_signature():
                return False
            else:
                return True